// by aqrit 
// v1 ( May 2, 2011)

#include <windows.h>
#include <Shlwapi.h>

#define text_len 160
#define text 	L"Select an executable that has 256-color related issues with DirectDraw.  " \
		L"Pressing 'Fix' will create a compatiblilty entry for this executable in the registry.\0\0"

void myError(HWND hWnd, char* caption){
	LPVOID lpMsgBuf;
	if( FormatMessage( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
		NULL, GetLastError(),MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR) &lpMsgBuf,0,NULL ))
	{
		MessageBox(hWnd, (char *)lpMsgBuf, caption, MB_OK | MB_ICONINFORMATION );
		LocalFree( lpMsgBuf );
	}
}

// returns a pointer to sub_str in str or null 
char* my_strstr ( const char* str, const char* sub_str ){
	size_t str_cnt     = lstrlen(str);
	size_t sub_str_cnt = lstrlen(sub_str);
	for(size_t i = 0; ((i + sub_str_cnt) <= str_cnt); i++){
		for(size_t ii = 0; ii < sub_str_cnt; ii++){
			if( str[i + ii] != sub_str[ii] )   
				break;
			if( ii >= sub_str_cnt - 1) // found
				return (char*)&str[i];
		}
	}
	return NULL; // not found
}


// some tricky alignment problems 
#include <pshpack2.h>
struct dialog_template_helper{ 
	DLGTEMPLATE dt; // must be DWORD aligned
	WORD menu;         // ordinal of a menu resource
	WORD windowClass;  // ordinal of a window class
	WCHAR title; // just NULL
	struct{
		DLGITEMTEMPLATE dit; // must be DWORD aligned
		WORD magic; //  -1 so we can use ordWndClass instead of szWndClass
		WORD ordWndClass; // ordinal value of a predefined system class
		WCHAR title[ text_len ]; // must be a multipul of 4 to maintain alignment
		WORD creationData; // length prefixed user data 
	}item_01;
};
#include <poppack.h>

dialog_template_helper dlg_data = 
{
	{ DS_3DLOOK | DS_CONTROL | WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS, 0, 1, 0, 0, 200, 50 },
	0, 0, 0,
	{ // item_01
		{  WS_CHILD | WS_VISIBLE, 0, 10, 10, 180, 40, 1 },
		-1,
		0x0082, // static
		L"Select an executable that has 256-color related issues with DirectDraw.  "
		L"Pressing 'Fix' will create a compatiblilty entry for this executable in the registry."
		L"\0", // extra null byte to pad for alignment
		0
	}
};

// Microsoft's User Account Control blocks wrtie access to HKEY_LOCAL_MACHINE
// administrator rights required
//
// sometimes things get redirected to the 'virtualstore' but it works from there as well so I'm not going to worry about it...
void CheckUAC(){
	HKEY hKey;
	if( RegOpenKeyEx( HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\DirectDraw\\Compatibility", 0, KEY_WRITE, &hKey ) != ERROR_SUCCESS){
	
		char * szCmdLine = GetCommandLine();
		
		if( my_strstr(szCmdLine, "-norestart") ){
			MessageBox(NULL,"Error: unable to obtain write access for 'HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\DirectDraw\\Compatibility'."
				"Make sure you you are running this program with administrator rights.","DDraw Color Fixer",MB_OK);
			ExitProcess(0);
		}else{
			char szFilePath[MAX_PATH];
			GetModuleFileName( GetModuleHandle(NULL), szFilePath, MAX_PATH );
			ShellExecute(NULL, "runas", szFilePath, "-norestart", NULL, SW_SHOWNORMAL); // runas ... admin
			ExitProcess(0);
		}
	}
	RegCloseKey(hKey);
}

void CreateCompatEntry(HWND hWnd, char* szFullPath){
	HANDLE hFile;
	char* szFileName;
	char* szFileExt;
	char szNameOnly[_MAX_FNAME];
	char szNameDecorated[_MAX_FNAME];
	DWORD id;

	if(szFullPath == (szFileName = PathFindFileName(szFullPath))){
		MessageBox(hWnd, "Unable to find the name from the path", "Error", MB_OK);
		return;
	}
	szFileExt = PathFindExtension( szFullPath );
	lstrcpy( szNameOnly, szFileName );
	szNameOnly[ lstrlen(szFileName) - lstrlen(szFileExt) ] = '\0'; 
	
	hFile = CreateFile(szFullPath, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	if(hFile == INVALID_HANDLE_VALUE){
		myError(hWnd, "Error Opening File");
		return;
	}
	OVERLAPPED ol = { 0,0,0,0 };
	DWORD dwBytesRead;
	char DosHeader[ sizeof(IMAGE_DOS_HEADER) ];
	char NtHeader[ sizeof(IMAGE_NT_HEADERS) ];
	
	// read-in dos header
	if(! ReadFile(hFile, DosHeader, sizeof(IMAGE_DOS_HEADER), &dwBytesRead, &ol)){
		myError(hWnd, "Error Reading File");
		CloseHandle(hFile);
		return;
	}

	PIMAGE_DOS_HEADER dos = (PIMAGE_DOS_HEADER) DosHeader;
	if(dwBytesRead < sizeof(IMAGE_DOS_HEADER) || dos->e_magic != IMAGE_DOS_SIGNATURE){
		MessageBox( hWnd,"Invalid Executable Signature", "Error", MB_OK );
		CloseHandle(hFile);
		return;
	}

	// read-in nt header
	ol.Offset = dos->e_lfanew;
	if(! ReadFile(hFile, NtHeader, sizeof(IMAGE_NT_HEADERS), &dwBytesRead, &ol)){
		myError(hWnd, "Error Reading File");
		CloseHandle(hFile);
		return;
	}

	PIMAGE_NT_HEADERS pe = (PIMAGE_NT_HEADERS) NtHeader;
	if( dwBytesRead < sizeof(IMAGE_NT_HEADERS) || pe->Signature != IMAGE_NT_SIGNATURE){
		MessageBox( hWnd,"Invalid NT Executable Signature", "Error", MB_OK );
		CloseHandle(hFile);
		return;
	}

	CloseHandle(hFile);
	
	id = pe->FileHeader.TimeDateStamp;

	HKEY hKey;
	HKEY hSubKey;
	LONG lResult;
	lResult = RegOpenKeyEx( HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\DirectDraw\\Compatibility", 0, KEY_WRITE, &hKey );
	if(lResult != ERROR_SUCCESS){
		MessageBox( hWnd, "Failed to open the registry key 'HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\DirectDraw\\Compatibility'.  "
			"Make sure the key exists and that you are running this program as an administrator", "Error", MB_OK);
		return;
	}

	wsprintf( szNameDecorated, "%s_%08X", szNameOnly, id );

	DWORD dwDisposition;
	lResult = RegCreateKeyEx(hKey, szNameDecorated, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_SET_VALUE, NULL, &hSubKey, &dwDisposition);
	if(lResult != ERROR_SUCCESS){
		MessageBox( hWnd, "Failed to create/open subkey", "Error", MB_OK);
		RegCloseKey(hKey);
		return;
	}
	if( REG_OPENED_EXISTING_KEY == dwDisposition ){
		// this name doesn't matter but to keep it simple just replace instead of spinning a off new name
		if( IDNO == MessageBox(hWnd, "Replace existing entry for this executable?", szNameDecorated, MB_YESNO )){
			RegCloseKey(hSubKey);
			RegCloseKey(hKey);
			return;
		}
	}

	DWORD Flags = 0x00000800; // fix 256-color problem flag
	lResult |= RegSetValueEx(hSubKey, "Name",  0, REG_SZ,    (BYTE*)szFileName, lstrlen(szFileName));
	lResult |= RegSetValueEx(hSubKey, "ID",    0, REG_DWORD, (BYTE*)&id,        sizeof(DWORD));
	lResult |= RegSetValueEx(hSubKey, "Flags", 0, REG_DWORD, (BYTE*)&Flags,     sizeof(DWORD));

	if( lResult == ERROR_SUCCESS ){
		MessageBox( hWnd, "Entry created successfully.", szNameOnly, MB_OK);
	}
	else{
		MessageBox( hWnd, "An error occured while setting subkey data", "Error", MB_OK);
	}

	RegCloseKey(hSubKey);
	RegCloseKey(hKey);
}


UINT CALLBACK OFNHookProc( HWND hdlg, UINT uiMsg, WPARAM wParam, LPARAM lParam ){
	if( uiMsg == WM_NOTIFY){
		switch( ((OFNOTIFY *)lParam)->hdr.code ){
			case CDN_FILEOK:
			{
				HWND hdlgParent = GetParent(hdlg);
				char szFullPath[_MAX_PATH];
				if( SendMessage(hdlgParent, CDM_GETFILEPATH, MAX_PATH, (LPARAM)szFullPath) > 0){
					CreateCompatEntry(hdlgParent, szFullPath);
				}else{
					MessageBox( hdlgParent, "Path of selection unknown", "Error", MB_OK );
				}

				// prevent exiting from GetOpenFileName dialog
				SetWindowLong( hdlg, DWL_MSGRESULT, TRUE );
				return TRUE;
			}
				//DWL_MSGRESULT
			case CDN_INITDONE:
			{
				HWND hdlgParent = GetParent(hdlg);
				// change button text
				SendMessage( hdlgParent, CDM_SETCONTROLTEXT, IDOK, (LPARAM)"Fix" ); 
				SendMessage( hdlgParent, CDM_SETCONTROLTEXT, IDCANCEL, (LPARAM)"Exit" ); 
				return TRUE;
			}
		}
	}
	return 0;
}


void __cdecl myStart(void){

	CheckUAC();

	DWORD dummy = 0;  
	OPENFILENAME ofn;
	HINSTANCE hInstance;

	hInstance = GetModuleHandle( NULL );
	RtlSecureZeroMemory(&ofn, sizeof(ofn)); // inline
	ofn.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_ENABLEHOOK | OFN_ENABLETEMPLATEHANDLE; //OFN_ENABLETEMPLATE;
	ofn.hInstance = (HINSTANCE) &dlg_data;
	ofn.lStructSize = sizeof(ofn);
	ofn.lpstrFilter = "Executable (*.exe)\0*.exe\0All Files\0*\0";
	ofn.lpstrFile = (LPSTR) &dummy;
	ofn.nMaxFile = sizeof(dummy) / sizeof(char); 
	ofn.lpstrTitle = "DDraw Color Fixer";
	ofn.lpfnHook = &OFNHookProc;
	GetOpenFileName( &ofn );

	ExitProcess(0);
}

